#include <windows.h>

#include <assert.h>
#include <stdlib.h>

#include "saohook/sao-dll.h"

#include "util/dll-bind.h"
#include "util/dprintf.h"

const struct dll_bind_sym sao_dll_syms[] = {
    {
        .sym = "sao_io_init",
        .off = offsetof(struct sao_dll, init),
    }, {
        .sym = "sao_io_read_coin_counter",
        .off = offsetof(struct sao_dll, read_coin_counter),
    }, {
        .sym = "sao_io_get_opbtns",
        .off = offsetof(struct sao_dll, get_opbtns),
    }, {
        .sym = "sao_io_get_drum_analog",
        .off = offsetof(struct sao_dll, get_drum_analog),
    }
};

struct sao_dll sao_dll;

HRESULT sao_dll_init(const struct sao_dll_config *cfg, HINSTANCE self)
{
    uint16_t (*get_api_version)(void);
    const struct dll_bind_sym *sym;
    HINSTANCE owned;
    HINSTANCE src;
    HRESULT hr;

    assert(cfg != NULL);
    assert(self != NULL);

    if (cfg->path[0] != L'\0') {
        owned = LoadLibraryW(cfg->path);

        if (owned == NULL) {
            hr = HRESULT_FROM_WIN32(GetLastError());
            dprintf("Sao IO: Failed to load IO DLL: %lx: %S\n",
                    hr,
                    cfg->path);

            goto end;
        }

        dprintf("Sao IO: Using custom IO DLL: %S\n", cfg->path);
        src = owned;
    } else {
        owned = NULL;
        src = self;
    }

    get_api_version = (void *) GetProcAddress(src, "sao_io_get_api_version");

    if (get_api_version != NULL) {
        sao_dll.api_version = get_api_version();
    } else {
        sao_dll.api_version = 0x0100;
        dprintf("Custom IO DLL does not expose sao_io_get_api_version, "
                "assuming API version 1.0.\n"
                "Please ask the developer to update their DLL.\n");
    }

    if (sao_dll.api_version >= 0x0200) {
        hr = E_NOTIMPL;
        dprintf("Sao IO: Custom IO DLL implements an unsupported "
                "API version (%#04x). Please update Segatools.\n",
                sao_dll.api_version);

        goto end;
    }

    sym = sao_dll_syms;
    hr = dll_bind(&sao_dll, src, &sym, _countof(sao_dll_syms));

    if (FAILED(hr)) {
        if (src != self) {
            dprintf("Sao IO: Custom IO DLL does not provide function "
                    "\"%s\". Please contact your IO DLL's developer for "
                    "further assistance.\n",
                    sym->sym);

            goto end;
        } else {
            dprintf("Internal error: could not reflect \"%s\"\n", sym->sym);
        }
    }

    owned = NULL;

end:
    if (owned != NULL) {
        FreeLibrary(owned);
    }

    return hr;
}